/*
 * Choice JS
*/

class Builder {

    //Constructor
    constructor(target, userOptions){
        this.target = target;
        //Build our default options
        this.options = {
            ...this.buildDefaultOptions(userOptions),
            ...userOptions
        };
    }
    
    
    /**
     * Create a new choice dropdown
     * 
     * @return {Promise}
    */
    async create(){
        //Build it
        let choice = await this.build();
        //Toast bindings
        await this.bindings(choice);
        //Deploy the select
        this.target.parentNode.insertBefore(choice, this.target.nextSibling);
        //Dispatch event to load it
        this.target.dispatchEvent(new Event('change'));
        //Done
        return choice;
    }
        
    
    /**
     * Build the default options
     * 
     * @param {Object} userOptions The user options for any overwrites
     * @return {Object} An object of options
    */
    buildDefaultOptions(userOptions){
        let options = {
            width:'inherit',
            height:'inherit',
            placeholder: (this.target.hasAttribute('placeholder') ? this.target.getAttribute('placeholder') : false),
            class: '',
            closeOnSelect: false,
            event: 'change',
            copyAttributes: true
        }
        //Return
        return options;
    }
    
    
    /**
     * Configure options for the choice
     * 
     * @param {HTMLElement} choice - The select we're configuring options for
     * @return {Promise}
    */
    async configureOptions(choice){
        
        //Width and height
        choice.style.width = (this.options.width == 'inherit') ? (this.target.offsetWidth + 'px') : this.options.width,
        choice.style.minHeight = (this.options.height == 'inherit') ? (this.target.offsetHeight + 'px') : this.options.height;
        
        //Class
        if(this.options.class != '') choice.classList.add(this.options.class);
    
        //Placeholder
        if(this.options.placeholder) choice.querySelector('.choicejs-select').innerHTML = this.options.placeholder;
        
        //Return
        return;
    }
    
    
    /**
     * Add option
     *
     * @param {HTMLElement} option - The option we're selecting
     * @returns {Promise}
    */
    async addOption(option){
        //Get elems
        let value = option.getAttribute('value'),
            multiple = (this.target.hasAttribute('multiple') && ['true', 'multiple'].includes(this.target.getAttribute('multiple')));
        //Mark the option as selected
        this.target.querySelector(`option[value="${value}"]`).selected = true;
        //If its not multiple, set the value of the parent select box
        if(!multiple){
            this.target.value = value;
        }
    }
    
    
    /**
     * Remove option
     *
     * @param {HTMLElement} option - The option we're selecting
     * @returns {Promise}
    */
    async removeOption(option){
        //Get elems
        let value = option.getAttribute('value'),
            multiple = (this.target.hasAttribute('multiple') && ['true', 'multiple'].includes(this.target.getAttribute('multiple')));
        //Remove the hidden option
        this.target.querySelector(`option[value="${value}"]`).selected = false;
        this.target.querySelector(`option[value="${value}"]`).removeAttribute('selected');
        //Single or multiple select?
        if(multiple){
            //If there is still options selected, return now, don't reset the value and index of the select
            if(this.target.querySelectorAll(`option:checked`).length > 0) return;
        }
        //Reset
        this.target.value = '';
        this.target.selectedIndex = -1;
    }
    
    
    /* ------------------------------------------------------------------------------------------------------------------------------------------------
    |  HTML Builders
    ------------------------------------------------------------------------------------------------------------------------------------------------ */
    
    
    /**
     * Build the choice
     * 
     * @return {Promise} HTMLElement The new choice
    */
    async build(){
        //Build the new Select HTML element and add it directly after the target
        let wrapper = document.createElement('div'),
            select = document.createElement('div'),
            search = document.createElement('input'),
            dropdown = document.createElement('div'),
            list = document.createElement('ul');
        //Set attributes
        wrapper.classList.add('choicejs');
        wrapper.setAttribute('id', 'choice-'+Math.random().toString(36).substr(2, 4));
        select.classList.add('choicejs-select');
        search.classList.add('choicejs-search');
        search.setAttribute('placeholder', 'Search');
        dropdown.classList.add('choicejs-dropdown');
        //Set the id on the target so we can relate
        this.target.setAttribute('data-choicejs-id', `#${wrapper.id}`);
        //Loop the targets options and build our list
        for(let option of this.target.querySelectorAll('option')){
            if(option.value != '' && option.value != undefined){
                let item = document.createElement('li');
                item.innerHTML = option.innerHTML;
                //Copy all attributes over from the option
                if(this.options.copyAttributes === true){                    
                    for (let i = 0; i < option.attributes.length; i++){
                        item.setAttribute(option.attributes[i].nodeName, option.attributes[i].nodeValue);
                    }
                }
                list.appendChild(item);
            }
        }
        //Build the whole select and return it
        dropdown.appendChild(search);
        dropdown.appendChild(list);
        wrapper.appendChild(select);
        wrapper.appendChild(dropdown);
        //Configure the options for it
        await this.configureOptions(wrapper);
        //Return
        return wrapper;
    }
    
    
    /**
     * Build option pill
     *
     * @param {Node} option - The option selected
     * @param {Node} select - The select we're appending to
     * @returns {Promise}
    */
    async addOptionPill(option, select){
        let wrapper = document.createElement('div'),
            close = document.createElement('span'),
            value = option.getAttribute('value');
        //Attributes
        wrapper.classList.add('choicejs-pill');
        wrapper.setAttribute('value', value);
        wrapper.innerHTML = option.innerHTML;
        //Build it
        wrapper.appendChild(close);
        select.appendChild(wrapper);
        //Add the remove event listener
        close.addEventListener('click', async (e) => {
            //Run remove btn
            await this.removeOption(option);
            //What parent event to trigger?
            if(this.options.event !== false){
                this.target.dispatchEvent(new Event(this.options.event));
            }
        })
        //Return
        return;
    }
        
    
    /* ------------------------------------------------------------------------------------------------------------------------------------------------
    |  Bindings
    ------------------------------------------------------------------------------------------------------------------------------------------------ */
    
    
    /**
     * Bind Event Listener
     * 
     * @param {HTMLElement} choice - The choice we're binding
     * @return {Promise}
    */
    async bindEventListener(choice){
        //When an event is fired on the target, change our dropdowns
        this.target.addEventListener(this.options.event, async () => {
            //Rebuild the options and pills on the target change
            let multiple = (this.target.hasAttribute('multiple') && ['true', 'multiple'].includes(this.target.getAttribute('multiple'))),
                choice_select = choice.querySelector('.choicejs-select');
            //Reset the select placeholder
            choice_select.innerHTML = (this.options.placeholder ? this.options.placeholder : '');
            //Reset all of the options
            for(let option of choice.querySelectorAll(`.choicejs-dropdown ul li`)){
                option.removeAttribute('selected');
            }
            //Are we working with multiple?
            if(!multiple){
                let selected_value = this.target.value;
                //Does the select have a value?
                if(selected_value != ''){
                    let choice_option = choice.querySelector(`.choicejs-dropdown ul li[value="${selected_value}"]`);
                    //Select it
                    choice_option.setAttribute('selected', true);
                    choice_select.innerHTML = choice_option.innerHTML;
                }
            } else {
                let pills = [],
                    selected_options = this.target.querySelectorAll('option:checked');
                if(selected_options.length > 0){
                    choice_select.innerHTML = '';
                    for(let option of selected_options){
                        let choice_option = choice.querySelector(`.choicejs-dropdown ul li[value="${option.value}"]`);
                        choice_option.setAttribute('selected', true);
                        //Add a pillbox
                        pills.push(this.addOptionPill(choice_option, choice_select));
                    }
                }
                await Promise.all(pills);
            }
        });
    }
    
    
    
    /**
     * Bind active click
     *
     * @param {HTMLElement} choice - The choice we're binding
     * @return {Promise}
    */
    async bindActiveListener(choice){
        choice.addEventListener('click', () => {
            choice.classList.add('active');
        })
        document.addEventListener('click', function(event) {
            if (!choice.contains(event.target)) {
                choice.classList.remove('active');
            }
        });
        //Return
        return;
    }
    
    
    /**
     * Bind option select
     * 
     * @param {HTMLElement} choice - The choice we're binding
     * @return {Promise}
    */
    async bindOptionSelect(choice){
        //Loop and add
        for(let option of choice.querySelectorAll('ul li')){
            //Add the event listener
            option.addEventListener('click', async(e) => {
                //Stop Prop
                e.stopPropagation();
                e.stopImmediatePropagation();
                //Add or remove
                if(option.hasAttribute('selected')){
                    //Remove the option
                    await this.removeOption(option);
                } else {
                    //Add the option
                    await this.addOption(option);
                }
                //Close on select?
                if(this.options.closeOnSelect){
                    choice.classList.remove('active');
                }
                //What parent event to trigger?
                if(this.options.event !== false){
                    this.target.dispatchEvent(new Event(this.options.event));
                }
            });
        } 
        //Return
        return;
    }
    
    
    /**
     * Bind option filtering. When a user types into the search box
     * 
     * @param {HTMLElement} choice - The choice we're binding
     * @return {Promise}
    */
    async bindOptionFiltering(choice){
        //When the choice search bar changes
        var searchTimeout = 0,
            searchBar = choice.querySelector('.choicejs-search'),
            options = choice.querySelectorAll('ul li');
        //When the postcode is changed, test it
        searchBar.addEventListener('input', () => {
            //Clear the above timeout
            clearTimeout(searchTimeout);
            //Set a new timeout
            searchTimeout = setTimeout(() =>{
                //Loop all options
                for(let option of options){
                    if(option.innerHTML.toLowerCase().includes(searchBar.value.toLowerCase())){
                        option.classList.remove('hide');
                    } else {
                        option.classList.add('hide');
                    }
                }
            }, 500);
        });
        //Return
        return;
    }
    
    
    /**
     * Add all of the bindings for the choice
     * 
     * @param {HTMLElement} choice - The select we're building the bindings for
     * @return {Promise}
    */
    async bindings(choice){
        //Return
        return await Promise.all([
            this.bindActiveListener(choice),
            this.bindOptionSelect(choice),
            this.bindOptionFiltering(choice),
            this.bindEventListener(choice)
        ]);
    }
    

}





/**
 * Export the choice function to use
 *
 * @param {NodeElement} target - The target we're adding the choice to
 * @param {Object} options - The options
 * @returns {Promise}  - An instance of choice
 */
export const Choice = async (target, options) => {
    //Start a new Choice
    let newBuild = new Builder(target, options);
    //Run the build
    return await newBuild.create();
}